#ifdef __linux__

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/select.h>
#include <termios.h>
#include <memory.h>
#include <stdint.h>
#include <time.h>

#define stricmp strcasecmp

#define _sleep(T) sleep(.001*(T))
static struct termios orig;
static void reset_term() {
	tcsetattr(0, TCSANOW, &orig);
}

void Sleep(uint32_t ms) {
	struct timespec req, rem;
	req.tv_sec = ms/1000;
	req.tv_nsec = ms*1000000;
	nanosleep(&req, &rem);	
}

#include <unistd.h>
#include <sys/syscall.h>
#include <sys/types.h>
#ifdef SYS_gettid
#define gettid syscall(SYS_gettid);
#else
// SYS_gettid unavailable
#endif





#define SetConsoleTitle(name) printf("%c]0;%s%c", '\033', name, '\007')

int _kbhit() {
	static int once=1;
   	if(once) { once=0;
		struct termios term;
		tcgetattr(0, &orig);
		memcpy(&term, &orig, sizeof(term));
		atexit(reset_term);
		term.c_lflag&=~(ICANON|ECHO);
		tcsetattr(0, TCSANOW, &term);
		//setbuf(stdin, 0);
	}
	struct timeval tv = {0};
	fd_set fds;
	FD_ZERO(&fds);
	FD_SET(0, &fds);
	return select(1, &fds, 0, 0, &tv);
}

int _getch() {
	char c = 0;
	int ret = read(fileno(stdin), &c, 1);
	return c;
}

#elif _WIN32
#include <Windows.h>
#include <conio.h>
#endif

void SetCWD(char *rel) {
	char dir[260], div;
	size_t len;
#ifdef __linux__
	len = readlink("/proc/self/exe", dir, sizeof(dir)-1);
	div = '/'; 
#elif _WIN32
	len = GetModuleFileNameA(NULL, dir, sizeof(dir)-1);
	div = '\\';
#endif
	while(len>0)
	if(dir[--len]==div) {
		if(rel==0) dir[len] = 0; // use exe directory
		else strcpy(&dir[++len], rel); // relative path
		break;
	}
#ifdef __linux__
	chdir(dir);
#elif _WIN32
	SetCurrentDirectoryA(dir);
#endif
//	puts(dir);
}

static uint64_t nanos() {
  static uint64_t is_init = 0;
#if defined(__APPLE__)
    static mach_timebase_info_data_t info;
    if (0 == is_init) {
      mach_timebase_info(&info);
      is_init = 1;
    }
    uint64_t now;
    now = mach_absolute_time();
    now *= info.numer;
    now /= info.denom;
    return now;
#elif defined(__linux)
    static struct timespec linux_rate;
    if (0 == is_init) {
      clock_getres(CLOCK_REALTIME, &linux_rate);
      is_init = 1;
    }
    uint64_t now;
    struct timespec spec;
    clock_gettime(CLOCK_REALTIME, &spec);
    now = spec.tv_sec * 1.0e9 + spec.tv_nsec;
    return now;
#elif defined(_WIN32)
    static LARGE_INTEGER win_frequency;
    if (0 == is_init) {
      QueryPerformanceFrequency(&win_frequency);
      is_init = 1;
    }
    LARGE_INTEGER now;
    QueryPerformanceCounter(&now);
    return (uint64_t) ((1e9 * now.QuadPart)  / win_frequency.QuadPart);
#endif
}
