// The Xmas Demo 2021: Control code
// Authors: Julin and Corneliusen
// More info here: https://www.ignorantus.com/pages/xmasdemo2021/
// License: CC0 1.0 Universal (CC0 1.0) Public Domain Dedication license

#include <V73D/V7.h>
#ifdef __ANDROID__

#elif _WIN32
#define _USE_MATH_DEFINES
#include <V73D/V7W32.h>
#else
//#include <V73D/V7GLX.h>
#endif

#include <V73D/font/sdf_babylon5.h>
#include <V73D/font/sdf_fradmit.h>
#include <V73D/font/sdf_rabbit.h>
#include <V73D/font/sdf_conflict.h>

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>

#include "console.h"

#include "A7XM.h" // The XM player

#include "raytracer.h"

#include "moose.h"

#ifdef _WIN32
A7X *A7Run(char *data, int size)
{
	(void)data;
	(void)size;
	return NULL;
}
void A7Stop(A7X *a7x)
{
	(void)a7x;
}
void A7Seek(A7X *a7x, int pos, int row, int tick)
{
	(void)a7x;
	(void)pos;
	(void)row;
	(void)tick;
}
#endif

// Set by options
static bool xavier = false;
static bool fullscreen = true;
static bool vsync = true;

// Set to render to jpg files
#define SAVEPICS false
#ifdef _WIN32
#define SAVEPATH "C:\\Users\\ncorn\\xmas2021r\\out%05d.jpg"
#else
#define SAVEPATH "../render/out%05d.jpg"
#endif

// Enables rendering of all frames, even if timing is off
bool usefaketime = SAVEPICS;

#define RENDERW 1920
#define RENDERH 1080

worldinfo worldray;
gpu_tracer gpuray;

#define PIF ((float)PI)

V7X *ctx;
A7X *a7x;


vec3 pygopos  = {   3.25f, -1.15f, 0.0f };
vec3 cobrapos = {   0.00f, -8.75f, 1.0f };
vec3 maxpos   = {  12.25f, -5.35f, 0.0f };
vec3 bikepos  = { -12.25f, -5.17f, 0.0f };

#define KAITXTS 3

vec3 kaipos[KAITXTS] = {
	{ -9.0f,  3.0f-1.0f, 0.0f },
	{ -9.0f,  0.0f-1.0f, 0.0f },
	{ -9.0f, -3.0f-1.0f, 0.0f },
};

char *kaitexts[KAITXTS] = {
	"CODE  FIRST\n",
	"CODE  HARD\n",
	"NO\n"
};


Model *bg[2], *fg[2], *txt[4];
Model *kaitxt[KAITXTS];
Model *cobrac, *pygo, *bike, *v73dlogo, *max;
// Main shaders
Model *bulbs, *fvalley, *jquat, *madbulb;
// Bg shaders
Model *fire, *smoke, *clouds, *fnoise;

Texture *cobratex, *pytex, *biketex, *v73dtex, *maxtex;

int part, switchpart, codemax;
double pstart, pdiff, prest, pnext, pfad=2.;

typedef enum {
	PART_M00SE0,
	PART_M00SE1,
	PART_KAI,
	PART_BULBS,
	PART_FVALLEY,
	PART_QUAT,
	PART_MAD,
	PART_CREDITS,
	PART_SNAKEPIT,
	PART_NADA,
	BODYPARTS
} party;

typedef struct Demopart_t {
	party partid;
	float len;
	vec3 titlepos, textpos;
	char *fg, *bg;
	char *title, *text, *cred;
} Demopart;

Demopart demopart[] = {
	{	PART_BULBS, 15.f, {0.0f, 2.85f, 2.f}, {0.0f, 0.8f, 2.0f},
		"Bulbs", "Fire",
		"\x01""The Xmas Demo 2021\n\n"
		"\x01""Julin & Corneliusen Productions\n",
		"\x01""Running on Nvidia Jetson AGX Xavier\n\n"
		"\x01""All code written in C and GLSL.\n\n"
		"\x01""Introducing the V73D engine\n\n",
		"Code by Corneliusen\n"
		"Background effect based on \"301's Fire Shader - Remix 3\" by mu6k\n"
	},

	{	PART_M00SE0, 30.f, {0.0f, 2.5f, 2.f}, {0.0f, 2.0f, 2.0f},
		"Empty", "Empty",
		"\x01""Deploy the M00SE\n",
		"\x01""Powered by Inverse Kinematics\n",
		"\nV73D engine by Julin"
	},

	{	PART_FVALLEY, 50.f, {3.25f, 3.05f, 2.f}, {-6.0f, 1.f, 2.0f},
		"FValley", "Empty",
		"Fire Valley\n",
		"",
		"\nBased on \"Menger's Dream\" by jamiep\n"
	},

	{	PART_KAI, 11.f, {-4.7f, 1.4f, 2.f}, {-6.0f, 1.f, 2.0f},
		"Empty", "FNoise",
		"",
		"",
		"\nBackground effect based on \"Fractal Noise\" by mu6k"
	},

	{	PART_M00SE1, 20.f, {0.0f, 2.5f, 2.f}, {-3.6f, 2.0f, 2.0f},
		"Empty", "Empty",
		"\x01""Meanwhile... in Moosetown\n",
		"",
		"\nMoose reanimated from a design by Weta Workshop"
	},

	{	PART_QUAT, 40.f, {-2.25f, 3.05f, 2.f}, {-5.9f, .0f, 2.0f},
		"JQuat", "Smoke",
		"Julia Quaternions Resurrection",
		"",
		"Code by Corneliusen\n"
		"Background effect based on \"Lights in Smoke\" by ehj1"
	},

	{	PART_MAD, 60.f, {3.6f, 3.05f, 2.f}, {-6.0f, 1.f, 2.0f},
		"Mad", "Clouds",
		"Mad Bulb\n",
		"",
		"Based on \"Mandelbulb Deconstructed\" by morgan3d\n"
		"Background effect based on \"2D Clouds\" by drift"
	},

	{	PART_CREDITS, 25.0f, {0.00f, -7.0f, 2.0f}, {6.0f, 1.f, 2.0f},
		"Empty", "Smoke",
		"\x01""The Xmas Demo 2021\n",
		"",
		"\n\n"
		"\x01""by\n\n"
		"\x01""Sjur Julin\n"
		"\x01""Nils Liaaen Corneliusen\n"
		"\n\n\n"
		"\x01""Music\n\n"
		"\x01""\"March of the Moose\" by Sjur Julin\n"
		"\x01""Samples from the movie Chappie (2015)\n"
		"\n\n\n"
		"\x01""Bitmap Graphics\n\n"
		"\x01""V73D, Pygo, and Cobra C logos by Sjur Julin\n"
		"\x01""(Any resemblance to existing logos is purely coincidental)\n"
		"\x01""Velocipede and Max Headroom by unknown artists\n"
		"\n\n\n"
		"\x01""3D Objects\n\n"
		"\x01""All objects by Sjur Julin\n"
		"\x01""Robot modeled after majestik M00SE by Weta Workshop\n"
		"\n\n\n"
		"\x01""Cobra C Code\n\n"
		"\x01""V73D Engine by Sjur Julin\n"
		"\x01""Control code by Julin & Corneliusen\n"
		"\n\n\n"
		"\x01""Shader Code\n\n"
		"\x01""\"Recycled Raytracer\" by Corneliusen\n"
		"\x01""\"Julia Quaternions Resurrection\" by Corneliusen\n"
		"\n"
		"\x01""The following shaders from shadertoy.com have been modified and used:\n"
		"\n"
		"\x01""\"Fractal Noise\" by mu6k\n"
		"\x01""\"301's Fire Shader - Remix 3\" by mu6k\n"
		"\x01""\"Menger's dream\" by jamiep\n"
		"\x01""\"Lights in Smoke\" by ehj1\n"
		"\x01""\"Mandelbulb Deconstructed\" by morgan3d\n"
		"\x01""\"2D Clouds\" by drift\n"
		"\n\n\n"
		"\x01""Other Code and Tools\n\n"
		"\x01""C glTF library by jkuhlmann\n"
		"\x01""vmath library by monolifed\n"
		"\x01""XM Player by Romain \"Artefact2\" Dalmaso\n"
		"\x01""SDF font generator by lonesock\n"
		"\x01""Music tracked in FastTrackerII\n"
		"\n\n\n"
		"\x01""Get the source code here:\n"
		"\n"
		"\x01""www.ignorantus.com/xmasdemo/\n"
		"\n\n\n\n\n\n\n\n"
		"\x01""Real Programming\n"
		"\x01""A book about programming, programmers, programs, and pop culture\n"
		"\n"
		"\x01""www.ignorantus.com/real/\n"
		"\n\n\n\n\n\n\n\n"
		"\x01""In writing and in programming\n"
		"\x01""I express myself as a member of a higher community\n"
		"\x01""I have forgotten how to walk and speak\n"
		"\x01""I am on the way toward flying into the air, coding\n"
		"\x01""My very gestures express enchantment\n"
		"\x01""I feel myself a god\n"
		"\x01""Welcome to real programming\n"
	},

	{	PART_SNAKEPIT, 7.0f, {-6.025f, 3.05f, 2.f}, {-6.0f, 1.f, 2.0f},
		"Empty", "Smoke",
		"",
		"",
		"\n"
		"\x01""Python is 72 times slower than C. Go green! Use C.\n"
	},

	{	PART_NADA, 5.0f, {-6.025f, 3.05f, 2.f}, {-6.0f, 1.f, 2.0f},
		"Empty", "Empty",
		"",
		"",
		""
		""
	},
};

#define NPARTS (sizeof(demopart)/sizeof(Demopart))

int quit;

extern float *sndbuf;
extern int sndbufsize;

typedef struct GPUFValley_t {
	float starttime;
} GPUFValley;

GPUFValley gpufvalley;

typedef struct GPUQuat_t {
	v4 start_pos;
	v4 mu_pos;
	v4 obj_col;
	float rot_xz;
	float ax;
	float ay;
	float starttime;
} GPUQuat;

GPUQuat gpuquat;

typedef struct GPUMadBulb_t {
	float starttime;
	float pitch;
	float yaw;
	float zoom;
	float power;
} GPUMadBulb;

GPUMadBulb gpumadbulb;

// Quat positions
#define MU_POSES 6

static const v4 postable[MU_POSES] = {
	{ -0.125f, -0.256f,   0.847f,   0.0895f }, // doh
	{ -1.0f,    0.2f,     0.0f,     0.0f    }, // stick
	{ -0.2f,    0.8f,     0.0f,     0.0f    },
	{ -0.137f, -0.630f,  -0.475f,  -0.046f  },
	{ -0.450f, -0.447f,   0.181f,   0.306f  },
	{ -0.125f, -0.256f,   0.847f,   0.0895f }, // doh
};

static const float zoomtable[MU_POSES] = {
	15.0f,
	11.75f,
	10.75f,
	10.5f,
	12.5f,
	15.0f,
};

#define QUAT_ROT_TIME (16*60)
#define QUAT_MIX_TIME (4*60)
#define QUAT_HOLD_TIME (4*60)
#define QUAT_MU_TIME (QUAT_HOLD_TIME+QUAT_MIX_TIME)

const float bgsize = 70.f;
const float bgspeed = 35.f;

Model *NewBG(char *name, char *shdname, void *gpudata, size_t gpusize) {
	static int count;
	Model *m = V7NewMod(ctx, name);
	V7Quad(m->msh, 16.f, 9.f);
	m->s = (vec3){7.f,7.f,7.f};
	m->t = (vec3){.0f,.0f,-22.f-count*.1f};
	char *shdsrc = V7Load(0, true, "../shaders/%s", shdname);
	if( xavier ) {
		char *ptr = strstr( shdsrc, "//#define XAVIER" );
		if( ptr ) {
			ptr[0] = ' ';
			ptr[1] = ' ';
		}
	}
	m->prg = V7CustPrg(ctx, name, shdsrc, shdname);
	if(gpudata) V7UniBuf(m, gpudata, gpusize);
//	m->flags|=V7NDEP;
	count++;
	return m;
}

Model *LoadMod(char *name) {
	Model *m = V7GetMod(ctx, name);
	if(m) return m;
	char fn[64];
	sprintf(fn, "../data/%s.glb", name);
	size_t glb_size;
	void *glb = V7Load(&glb_size, 0, fn);
	if(!glb) {
		V7Error("Can't load model '%s'", fn);
		return 0;
	}
	m = V7GLB(ctx, name, glb, glb_size);
	m->Update = 0;
	return m;
}

static Model *LoadPNGAsMod( Scene *s, char *fname, char *name, Texture **tex, vec3 sc, vec3 t )
{
	char sname[80];

	void *pic = V7Load(NULL, true, fname );

	snprintf( sname, sizeof(sname), "%sTex", name );
	*tex = V7NewTex(ctx, sname);
	snprintf( sname, sizeof(sname), "%sImg", name );
	(*tex)->img = V7NewImg(ctx, sname);
	V7PNGDec((*tex)->img, pic);
	V7ImgTex(*tex);

	snprintf( sname, sizeof(sname), "%sMod", name );
	Model *mod = V7NewMod(ctx, sname);
	mod->prg = V7GetPrg(ctx, "RGB");
	V7Quad(mod->msh, 2.0f*((*tex)->img->w/(float)(*tex)->img->h), 2.0f);
	snprintf( sname, sizeof(sname), "%sMtr", name );
	mod->mtr = V7NewMtr(ctx, sname);
	mod->mtr->coltex = *tex;
	mod->s = sc;
	mod->t = t;

	V7AddMod(s, mod);

	V7Hide( mod );

	return mod;
}

static Model *NewText( char *title, SDFont *fn, vec3 sc, vec3 t, vec4 col )
{
	Model *mo = V7NewTxt( ctx, title, fn );
	mo->s = sc;
	mo->t = t;
	mo->mtr->col = col;
	V7AddMod( ctx->scn, mo );
	return mo;
}


bool mute = false;

double faketime = 0.0;
int framectr = 0;

void Setup(void) {

	Scene *s = ctx->scn = V7NewScn(ctx, "MainScene");
	s->dat.light[0].pos.x = -3.f;
	s->dat.light[0].pos.y =  6.f;
	s->dat.light[0].pos.z =  9.f;

	for(int i=0; i<2; i++) {
		bg[i] = V7NewMod(ctx, "BG");
		V7AddMod(s, bg[i]);
	}

	MooseInit(ctx);

	world_init(  &worldray, &gpuray, RENDERW, RENDERH );

	(void)LoadMod("Empty");

	bulbs    = NewBG("Bulbs",   "bulbs.fs",     &gpuray,     sizeof(gpuray));
	fvalley  = NewBG("FValley", "firevalley.fs",&gpufvalley, sizeof(gpufvalley));
	jquat    = NewBG("JQuat",   "quat.fs",      &gpuquat,    sizeof(gpuquat));
	madbulb  = NewBG("Mad",     "madbulb.fs",   &gpumadbulb, sizeof(gpumadbulb));
	fire     = NewBG("Fire",    "fire.fs",      0,           0); // bg for bulbs
	fnoise   = NewBG("FNoise",  "fracnoise.fs", 0,           0); // bg for kai
	smoke    = NewBG("Smoke",   "smoke.fs",     0,           0); // bg for quat, credits, snakepit
	clouds   = NewBG("Clouds",  "clouds.fs",    0,           0); // bg for madbulb

	for(int i=0; i<2; i++) {
		fg[i] = V7NewMod(ctx, "FG");
		V7AddMod(s, fg[i]);
	}

	SDFont *fnt[4];
	fnt[0] = V7NewFnt(ctx, &sdf_babylon5);
	fnt[1] = V7NewFnt(ctx, &sdf_fradmit);
	fnt[2] = V7NewFnt(ctx, &sdf_conflict); // Cobra font
	fnt[3] = V7NewFnt(ctx, &sdf_rabbit);   // Debug font

	txt[0] = NewText( "TitleTxt", fnt[0], (vec3){.5f,.5f,.5f}, (vec3){-3.f,     2.f,   2.f },(vec4){ 0.9f, 0.7f, 0.3f, 1.0f} );
	txt[1] = NewText( "CredTxt",  fnt[1], (vec3){.2f,.2f,.2f}, (vec3){-6.025f, -3.15f, 2.0f},(vec4){ 0.0f, 1.0f, 1.0f, 1.0f} );
	txt[2] = NewText( "InfoTxt",  fnt[1], (vec3){.4f,.4f,.4f}, (vec3){-2.7f,   -1.f,   2.0f},(vec4){ 0.0f, 1.0f, 1.0f, 1.0f} );
	txt[3] = NewText( "DebugTxt", fnt[3], (vec3){.1f,.1f,.1f}, (vec3){ 1.5f,   -2.75f, 2.f },(vec4){ 0.0f, 1.0f, 1.0f, 1.0f} );

	txt[2]->mtr = V7NewMtr( ctx, "weeaboo" );
	*txt[2]->mtr = *txt[1]->mtr;
	txt[2]->mtr->col = (vec4) { 1.f, 1.f, 1.f, 1.f};

	V7Hide( txt[2] );
	V7Hide( txt[3] );

	for( int i = 0; i < KAITXTS; i++ ) {
		char buf[80];
		snprintf( buf, sizeof(buf), "KaiTxt%d", i );
		kaitxt[i] = NewText( buf, fnt[2], (vec3) {4.75f, 4.75f, 4.75f}, kaipos[i], (vec4) { 1.0f, 1.0f, 0.0f, 1.f} );
		V7Txt(kaitxt[i], kaitexts[i]);
		V7Hide(kaitxt[i]);
	}

	cobrac   = LoadPNGAsMod( s, "../data/cobra_c.png", "CobraC", &cobratex, (vec3){ 4.00f, 4.00f, 0.0f}, cobrapos );
	pygo     = LoadPNGAsMod( s, "../data/python.png",  "Pygo",   &pytex,    (vec3){ 3.75f, 3.75f, 0.0f}, pygopos );
	bike     = LoadPNGAsMod( s, "../data/bike.png",    "Bike",   &biketex,  (vec3){ 0.50f, 0.50f, 0.0f}, bikepos );
	v73dlogo = LoadPNGAsMod( s, "../data/logo.png",    "V73D",   &v73dtex,  (vec3){ 0.25f, 0.25f, 0.0f}, (vec3){ 9.07f, -5.39f, 0.0f} );
	max      = LoadPNGAsMod( s, "../data/max.png",     "Max",    &maxtex,   (vec3){ 0.33f, 0.33f, 0.0f}, maxpos );

	V7CompShd(ctx);

	pstart = usefaketime ? faketime : V7Time();
}

///////////////////////////////////////////////////////////////////// LOOP //
void Loop(void) {

	faketime = framectr/60.0;
	if( usefaketime ) ctx->dat.time = (float)faketime;

	static float credstarttime;

	if(ctx->key[_ESC_]) { // ESC
		quit=1;
		ctx->key[_ESC_]=0;
	}
#ifndef __ANDROID__
//	if(ctx->key[_R_]||ctx->dat.frame==60) {
	if(ctx->key[_R_]) {
		ctx->key[_R_]=0;
		V7ReloadShd(ctx);
		V7CompShd(ctx);
		V7List(ctx);
	}
#endif

	if(ctx->key[_L_]) {
		ctx->key[_L_] = 0;
		V7List(ctx);
	}
	if(ctx->key[_M_]) {
		ctx->key[_M_] = 0;
		V7PrtMtx(ctx->scn);
	}

	if(ctx->key[_SPACE_]) {
		ctx->key[_SPACE_] = 0;
		switchpart=part+1;
	}

	double t = usefaketime ? faketime : V7Time();

	pdiff = t-pstart;
	prest = pnext-t;

	if(pdiff > demopart[part].len)
		switchpart = part+1;

	static bool kaion = false;
	static float kaistart = 0.0f;
	static int kaistep = -1;

	txt[1]->r.y = 0.0f;

	if(switchpart!=-1) {
		part = switchpart%NPARTS;
		switchpart = -1;
		pstart = t;
		pnext = t + demopart[part].len;
		txt[0]->t = demopart[part].titlepos;
		V7Txt(txt[0], demopart[part].title);
		*bg[1] = *bg[0]; bg[1]->t.z = -21.9;
		*bg[0] = *V7GetMod(ctx, demopart[part].bg); bg[0]->t.z = -22.0f;
		*fg[1] = *fg[0]; fg[1]->t.z = -19.9;
		*fg[0] = *V7GetMod(ctx, demopart[part].fg); fg[0]->t.z = -20.0f;

		txt[1]->t = (vec3) {-6.025f, -3.15f, 2.0f};
		txt[1]->mtr->col = (vec4) { 0.0f, 1.0f, 1.0f, 1.0f };

		V7Hide( bike );
		V7Hide( v73dlogo );
		V7Hide( max );

		kaion = false;

		switch( demopart[part].partid ) {
		case PART_M00SE0:	MooseStart(ctx, 0); break;
		case PART_M00SE1:	MooseStart(ctx, 1); break;
		case PART_KAI:		kaion = true; kaistep = -1; break;
		case PART_SNAKEPIT:	txt[1]->mtr->col = (vec4) { 0.0f, 1.0f, 0.0f, 1.0f }; txt[1]->t = (vec3) {0.0f, -3.15f, 2.0f}; V7Show( cobrac ); break;
		case PART_BULBS:	world_init(  &worldray, &gpuray, RENDERW, RENDERH ); break;
		case PART_FVALLEY:	gpufvalley.starttime = ctx->dat.time; break;
		case PART_QUAT:		V7Show( max );  max->t  = maxpos;  gpuquat.starttime    = ctx->dat.time; break;
		case PART_MAD:		V7Show( bike ); bike->t = bikepos; gpumadbulb.starttime = ctx->dat.time; break;
		case PART_CREDITS:	credstarttime = ctx->dat.time; V7Show( v73dlogo ); txt[1]->t.x = 0.0f; break;
		}

		txt[2]->t = demopart[part].textpos;
		V7Txt(txt[2], demopart[part].text);
		codemax = (int)txt[2]->msh->vidx.num;
		txt[2]->msh->vidx.num = 0;
		V7Hide(txt[2]); // avoid flash on change
	} else {
		V7Show(txt[2]); // show after change
	}

	V7Txt(txt[1], "%s", demopart[part].cred);

	if( demopart[part].partid == PART_CREDITS ) {
		float ftime = ctx->dat.time-credstarttime;
		int frm = (int)roundf(ftime * 60.0f);
		txt[0]->t.y = demopart[part].titlepos.y + frm*0.020f;
		txt[1]->t.y = demopart[part].titlepos.y + frm*0.020f;
		txt[1]->r.y = sinf(ftime+1.0f)/20.0f;
	}
	if( demopart[part].partid == PART_MAD && ctx->dat.time-gpumadbulb.starttime >= 25.0f ) {
		float ftime = ctx->dat.time-gpumadbulb.starttime-25.0f;
		bike->t.x = bikepos.x + 0.075f*roundf(ftime*60.0f);
	}

	if( demopart[part].partid == PART_QUAT && ctx->dat.time-gpuquat.starttime >= 10.0f  ) {
		float ftime = ctx->dat.time-gpuquat.starttime-10.0f;
		if( ftime <= 5.0f ) {
			max->t.x = maxpos.x - sinf( ftime/5.0f*PIF )*maxpos.x;
		}
	}

	if( !kaion ) {
		for( int i = 0; i < KAITXTS; i++ ) {
			V7Hide( kaitxt[i] );
		}
		V7Hide( cobrac );
		V7Hide( pygo );
	} else {

		if( kaistep == -1 ) {
			pygo->t = pygopos;
			cobrac->t = cobrapos;
			V7Hide( cobrac );
			V7Hide( pygo );
			kaistart = ctx->dat.time;
			kaistep = 0;
			for(int i = 0; i < KAITXTS; i++ )
				kaitxt[i]->t = kaipos[i];
		}

		if( (ctx->dat.time-kaistart) >= 2.5f )   V7Show( kaitxt[0] );
		if( (ctx->dat.time-kaistart) >= 3.5f )   V7Show( kaitxt[1] );
		if( (ctx->dat.time-kaistart) >= 4.5f ) { V7Show( kaitxt[2] ); V7Show( pygo ); V7Show( cobrac ); }
		if( (ctx->dat.time-kaistart) >= 5.5f ) {
			txt[1]->t = (vec3) {0.0f, -3.15f, 2.0f};
			txt[1]->mtr->col = (vec4) { 0.0f, 1.0f, 0.0f, 1.0f };
			V7Txt( txt[1], "%s", "\n\x01""Python uses 76 times more energy than C. Go green! Use C.\n" );

			if( pygo->t.y < 10.5f ) {
				float ftime = ctx->dat.time-kaistart-5.5f;
				int fr = (int)roundf(ftime * 60.0f);
				for( int i = 0; i < KAITXTS; i++ ) {
					kaitxt[i]->t.y = kaipos[i].y + fr * 0.30f;
				}
				pygo->t.y = pygopos.y + fr * 0.30f;
			}

			if( cobrac->t.y < 0.0f ) {
				float ftime = ctx->dat.time-kaistart-5.5f;
				if( ftime <= 1.0f ) {
					cobrac->t.y = cobrapos.y - cobrapos.y*(sinf( ftime/2.0f*PIF/1.0f ));
				}
				if( cobrac->t.y > 0.0f ) cobrac->t.y = 0.0f;
			}
		}
	}

	if( demopart[part].partid == PART_SNAKEPIT ) {
		cobrac->t = (vec3){0.0f,0.0f,1.0f};
		V7Show( cobrac );
	}

	int nvtx = (int)(pdiff * 160);
	txt[2]->msh->vidx.num = nvtx > codemax ? codemax : nvtx < 0 ? 0 : nvtx;
	V7MeshUp(txt[2]->msh);

	float tdis = 1.0f;
	float bfdis = 1.0f;
	if( ctx->dat.time + pfad >= pnext ) {
		// fade text out during last 2 secs
		tdis = ((float)pnext - ctx->dat.time)/(float)pfad;
		bfdis = 1.0f;
	} else if( ctx->dat.time - pstart <= pfad ) {
		// fade text in and fg+bg out/in first 2 secs
		tdis = (ctx->dat.time - (float)pstart)/(float)pfad;
		bfdis = tdis;
	} else {
		// don't wrap demo
		if( demopart[part].partid == PART_NADA ) exit( 1 );
	}

	fg[0]->dissolve = bg[0]->dissolve = (1.0f-bfdis);
	fg[1]->dissolve = bg[1]->dissolve = bfdis;

	for( int i = 0; i < 4; i++ ) txt[i]->mtr->col.w = tdis;

	if( demopart[part].partid == PART_SNAKEPIT || demopart[part].partid == PART_KAI ) {
		cobrac->dissolve = sinf( PIF*0.5f + tdis*PIF*0.5f );
	}
	if( demopart[part].partid == PART_CREDITS ) {
		v73dlogo->dissolve = sinf( PIF*0.5f + tdis*PIF*0.5f );
	}

	float minpower = 2.0f;
	float maxpower = 8.0f;
	float powerlen = 33.5f;

	float zoommax = 1.9f;
	float zoomtime = 6.0f;

	float tt = ctx->dat.time - gpumadbulb.starttime;
	if( tt < zoomtime )
		gpumadbulb.zoom = (1.0f-cosf(tt/zoomtime*PIF/2.0f))*zoommax;
	else
		gpumadbulb.zoom = zoommax;

	gpumadbulb.pitch = ((sinf(tt/12.0f)+1.0f)/0.5f)*PIF; // up/down 0..2*PI
	gpumadbulb.yaw   = tt/2.0f;                         // left/right ditto
	gpumadbulb.power = minpower + (cosf(PIF+PIF*tt/powerlen)+1.0f)*0.5f *(maxpower-minpower);

	static int lastframe;

	int actframe = (int)roundf(ctx->dat.time*60.0f);
	if( lastframe == 0 ) lastframe = actframe-1;

	for( int i = 0; i < actframe-lastframe; i++ )
		world_timer( &worldray, &gpuray );
	world_probe( &worldray, &gpuray );

	lastframe = actframe;

	gpuquat.ax  = 10.0f * 16.0f / 9.0f;
	gpuquat.ay  = 10.0f;
	gpuquat.start_pos = v4_set( gpuquat.ax  * -0.5f - 0.0f, gpuquat.ay  * -0.5f - 0.0f, 12.0f, 0.0f );

#define OVERSHOOT 1.25f
#define ZOOMSHOOT 1.6f
	int frs = (int)(roundf((ctx->dat.time-gpuquat.starttime)*60.0f));
	int smu = (frs/QUAT_MU_TIME)%(MU_POSES-1);
	int qfr = frs%QUAT_MU_TIME;
	static v4 topmu;
	static float topz;
	if( qfr < QUAT_MIX_TIME/2 ) {
		float wt = sinf( (qfr%QUAT_MIX_TIME)/(float)(QUAT_MIX_TIME) * PIF );
		gpuquat.mu_pos = v4_mix( postable[smu], v4_mul4( postable[smu+1], OVERSHOOT), wt );
		gpuquat.start_pos.z = zoomtable[smu]*(1.0f-wt) + zoomtable[smu+1]*wt*ZOOMSHOOT;
		topmu = gpuquat.mu_pos;
		topz = gpuquat.start_pos.z;
	} else if( qfr < QUAT_MIX_TIME ) {
		float wt = (sinf( ( (((qfr-QUAT_MIX_TIME/2)%QUAT_MIX_TIME)/(float)QUAT_MIX_TIME) * (PIF*2.0f) + PIF/2.0f ))+ 1.0f)/2.0f;
		gpuquat.mu_pos = v4_mix( postable[smu+1], topmu, wt );
		gpuquat.start_pos.z = zoomtable[smu+1]*(1.0f-wt) + topz*(wt);
	} else {
		gpuquat.mu_pos = postable[smu+1];
		gpuquat.start_pos.z = zoomtable[smu+1];
	}

	int rfr = frs%QUAT_ROT_TIME;
	gpuquat.rot_xz = sinf( rfr/(float)(QUAT_ROT_TIME) * 2.0f*(PIF) ) * PIF;
	gpuquat.obj_col = h2rgb3( frs/(float)(QUAT_ROT_TIME*2) );

	if( demopart[part].partid == PART_M00SE0 || demopart[part].partid == PART_M00SE1 ) {
		MooseLoop(ctx);
	} else {
		MooseHide(ctx);
	}

	V7Render(ctx);

	if( SAVEPICS ) {
		char fn[256];
		snprintf( fn, sizeof(fn), SAVEPATH, ctx->dat.frame );
		puts(fn);
		V7GrabFrame( ctx, fn );
	}

	framectr++;
}

int main(int argc, char **argv) {

	for( int i = 1; i < argc; i++ ) {
		if( !strcmp( argv[i], "-h" ) ) {
			printf( "Usage: %s [-x] [-w]\n", argv[0] );
			printf( "  -x: Xavier mode\n" );
			printf( "  -w: Run in window\n" );
			printf( "  -s: Disable vsync\n" );
			return 0;
		}
		if( !strcmp( argv[i], "-v" ) ) {
			printf( "The Xmas Demo 2021 by Julin & Corneliusen\n" );
			printf( "More info here: https://www.ignorantus.com/xmasdemo/\n" );
			return 0;
		}
		if( !strcmp( argv[i], "-x" ) ) {
			xavier = true;
			printf( "Xavier mode enabled\n" );
			for( int j = 0; j < NPARTS; j++ ) {
				party p = demopart[j].partid;
				if( p != PART_KAI && p != PART_CREDITS && p != PART_SNAKEPIT )
					demopart[j].bg = "Empty";
			}
		}
		if( !strcmp( argv[i], "-w" ) ) {
			fullscreen = false;
			printf( "Windowed mode enabled\n" );
		}
		if( !strcmp( argv[i], "-s" ) ) {
			vsync = false;
			printf( "Vsync disabled\n" );
		}
	}

	SetCWD(0);

	ctx = V7Create(L"A M00se once bit my sister...", 0, 0, RENDERW, RENDERH, fullscreen, vsync?1:0);
	if(!ctx) return -1;
	if( usefaketime ) V7SET( ctx->flags, V7TOFF );

	// hack
	ctx->dat.width  = RENDERW;
	ctx->dat.height = RENDERH;
	ctx->dat.aspect = ctx->dat.width / (float)ctx->dat.height;

	time_t t;
	srand((unsigned) time(&t));

	size_t xmsize;
	char *xm = V7Load(&xmsize, 0, "../data/MEH-76.xm");
	a7x = A7Run(xm, (int)xmsize); // Start music thread

	Setup();

	while(!quit) {
		Loop();
	}

	printf("QUIT!");

	V7Destroy(ctx);

	A7Stop(a7x); // Stop music

	return 0;
}
